from typing import List, Tuple
from flwr.server import ServerApp, ServerConfig
from flwr.server.strategy import FedAvg, FedMedian, FedOpt, FedAdam
from flwr.common import Metrics, Parameters
import argparse
import tensorflow as tf
import flwr

# Create the argument parser
parser = argparse.ArgumentParser(description="Server setting for FLWR")

# Add the argument for num_rounds (-r or --rounds)
parser.add_argument("-r", "--rounds", type=int, default=10, help="Number of rounds")

# Add the argument for strategy (-s or --strategy)
parser.add_argument(
    "-s",
    "--strategy",
    type=str,
    default="fedAvg",
    choices=["fedAvg", "fedMedian", "fedOpt", "fedAdam"],  # Add options strategies
    help="Select strategy: fedAvg, fedMedian, fedOpt or fedAdam"
)

# Parse the arguments from the command line
args = parser.parse_args()

# Define metric aggregation function
def weighted_average(metrics: List[Tuple[int, Metrics]]) -> Metrics:
    # Multiply accuracy of each client by number of examples used
    accuracies = [num_examples * m["accuracy"] for num_examples, m in metrics]
    examples = [num_examples for num_examples, _ in metrics]

    # Aggregate and return custom metric (weighted average)
    return {"accuracy": sum(accuracies) / sum(examples)}

# Load model for server-side parameter initialization
model = tf.keras.applications.MobileNetV2(
    input_shape=(32, 32, 3), weights=None, classes=10
)
model.compile("adam", "sparse_categorical_crossentropy", metrics=["accuracy"])

# Get model weights as a list of NumPy ndarray's
weights = model.get_weights()

# Serialize ndarrays to `Parameters`
parameters = flwr.common.ndarrays_to_parameters(weights)

# Select the strategy based on the input argument
if args.strategy == "fedAvg":
    strategy = FedAvg(evaluate_metrics_aggregation_fn=weighted_average)
elif args.strategy == "fedMedian":
    strategy = FedMedian(evaluate_metrics_aggregation_fn=weighted_average)
elif args.strategy == "fedOpt":
    strategy = FedOpt(evaluate_metrics_aggregation_fn=weighted_average, initial_parameters=parameters)
elif args.strategy == "fedAdam":
    strategy = FedAdam(evaluate_metrics_aggregation_fn=weighted_average, initial_parameters=parameters)

# Define config
config = ServerConfig(num_rounds=args.rounds)


# # Flower ServerApp
# app = ServerApp(
#     config=config,
#     strategy=strategy,
# )


# Legacy mode
if __name__ == "__main__":
    from flwr.server import start_server

    start_server(
        server_address="0.0.0.0:8080",
        config=config,
        strategy=strategy,
    )
