import csv
import time
import os
from datetime import datetime
from typing import List, Tuple, Dict, Optional
from flwr.common import Parameters, Metrics
from flwr.server.strategy import Strategy
from flwr.server.client_manager import ClientManager
from flwr.server.client_proxy import ClientProxy
from flwr.server import ServerApp, ServerConfig
from flwr.server.strategy import FedAvg, FedMedian, FedOpt, FedAdam, FedProx, QFedAvg
import argparse
import tensorflow as tf
import flwr

# Create the folder and generate the CSV filename dynamically
def create_csv_filepath(strategy_name: str, num_rounds: int) -> str:
    # Determine the parent directory of the current script
    parent_dir = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

    # # Create the "Log_autres" directory if it doesn't exist
    # log_dir = os.path.join(parent_dir, "Log_autres")
    # os.makedirs(log_dir, exist_ok=True)
    
    log_dir = args.log_dir
    if log_dir is None:
        log_dir = os.path.join(parent_dir, "Log_autres")
        os.makedirs(log_dir, exist_ok=True)

    # Generate a timestamp for the file
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")

    # Generate the CSV file path
    csv_filename = f"training_results_{strategy_name}_{num_rounds}_{timestamp}.csv"
    csv_filepath = os.path.join(log_dir, csv_filename)

    return csv_filepath

class CustomStrategyWrapper(Strategy):
    def __init__(self, strategy: Strategy, csv_filename: str):
        self.strategy = strategy
        self.csv_filename = csv_filename

        # Initialize CSV file and write header
        with open(self.csv_filename, mode="w", newline="") as f:
            writer = csv.writer(f)
            writer.writerow(["round", "time", "accuracy"])  # Header

    def initialize_parameters(self, client_manager: ClientManager) -> Optional[Parameters]:
        return self.strategy.initialize_parameters(client_manager)

    def configure_fit(
        self, server_round: int, parameters: Parameters, client_manager: ClientManager
    ):
        # Log the round start time
        self.round_start_time = time.time()
        return self.strategy.configure_fit(server_round, parameters, client_manager)

    def aggregate_fit(
        self,
        server_round: int,
        results: List[Tuple[ClientProxy, Metrics]],
        failures: List[BaseException],
    ):
        # Call the actual strategy's aggregate_fit
        return self.strategy.aggregate_fit(server_round, results, failures)

    def configure_evaluate(
        self, server_round: int, parameters: Parameters, client_manager: ClientManager
    ):
        return self.strategy.configure_evaluate(server_round, parameters, client_manager)

    def aggregate_evaluate(
        self,
        server_round: int,
        results: List[Tuple[ClientProxy, Metrics]],
        failures: List[BaseException],
    ):
        # Call the actual strategy's aggregate_evaluate
        aggregated_metrics = self.strategy.aggregate_evaluate(server_round, results, failures)

        # Calculate round duration
        round_time = time.time() - self.round_start_time

        # Extract accuracy from aggregated_metrics
        accuracy = aggregated_metrics[1].get("accuracy", 0) if aggregated_metrics else 0

        # Append data to CSV
        with open(self.csv_filename, mode="a", newline="") as f:
            writer = csv.writer(f)
            writer.writerow([server_round, round_time, accuracy])

        return aggregated_metrics

    def evaluate(
        self, server_round: int, parameters: Parameters
    ) -> Optional[Tuple[float, Metrics]]:
        return self.strategy.evaluate(server_round, parameters)


# Create the argument parser
parser = argparse.ArgumentParser(description="Server setting for FLWR")

# Add the argument for num_rounds (-r or --rounds)
parser.add_argument("-r", "--rounds", type=int, default=10, help="Number of rounds")

# Add the argument for strategy (-s or --strategy)
parser.add_argument(
    "-s",
    "--strategy",
    type=str,
    default="fedAvg",
    choices=["fedAvg", "fedMedian", "fedOpt", "fedAdam", "fedAvg2Clients", "fedProx", "QfedAvg"],  # Add options strategies
    help="Select strategy: fedAvg, fedMedian, fedOpt, fedAdam, fedProx, QfedAvg or fedAvg2Clients",
)

parser.add_argument("-l", "--log_dir", type=str, default=None, help="Directory to save logs")

# Parse the arguments from the command line
args = parser.parse_args()

# Define metric aggregation function
def weighted_average(metrics: List[Tuple[int, Metrics]]) -> Metrics:
    # Multiply accuracy of each client by number of examples used
    accuracies = [num_examples * m["accuracy"] for num_examples, m in metrics]
    examples = [num_examples for num_examples, _ in metrics]

    # Aggregate and return custom metric (weighted average)
    return {"accuracy": sum(accuracies) / sum(examples)}

# Load model for server-side parameter initialization
model = tf.keras.applications.MobileNetV2(
    input_shape=(32, 32, 3), weights=None, classes=10
)
model.compile("adam", "sparse_categorical_crossentropy", metrics=["accuracy"])

# Get model weights as a list of NumPy ndarray's
weights = model.get_weights()

# Serialize ndarrays to `Parameters`
parameters = flwr.common.ndarrays_to_parameters(weights)

# Select the base strategy
if args.strategy == "fedAvg": 
    # baseline strategy
    base_strategy = FedAvg(
        evaluate_metrics_aggregation_fn=weighted_average)
elif args.strategy == "fedAvg2Clients": 
    # Custom client num each round
    base_strategy = FedAvg(
        fraction_fit=0.67, # 67% of clients agg/fit each round
        min_fit_clients=2, # At least 2 clients fit each round
        min_available_clients=2, # At least 2 clients available each round
        evaluate_metrics_aggregation_fn=weighted_average,
    )
elif args.strategy == "fedMedian":
    # robustness strategy
    base_strategy = FedMedian(
        evaluate_metrics_aggregation_fn=weighted_average)
elif args.strategy == "fedOpt": 
    # server optimization strategy
    base_strategy = FedOpt(
        evaluate_metrics_aggregation_fn=weighted_average, 
        initial_parameters=parameters
    )
elif args.strategy == "fedAdam":
    # server optimization strategy use Adam
    base_strategy = FedAdam(
        evaluate_metrics_aggregation_fn=weighted_average, 
        initial_parameters=parameters
    )
elif args.strategy == "fedProx":
    # regularization strategy
    base_strategy = FedProx(
        evaluate_metrics_aggregation_fn=weighted_average, 
        initial_parameters=parameters,
        proximal_mu = 0.1
    )
elif args.strategy == "QFedAvg":
    # improve fairness, reduce var_loss strategy
    base_strategy = QFedAvg(
        evaluate_metrics_aggregation_fn=weighted_average, 
        initial_parameters=parameters
    )


# Generate the CSV file path based on strategy and number of rounds
csv_filepath = create_csv_filepath(strategy_name=args.strategy, num_rounds=args.rounds)

# Wrap the selected strategy with the logging wrapper
strategy = CustomStrategyWrapper(strategy=base_strategy, csv_filename=csv_filepath)

# Define config
config = ServerConfig(num_rounds=args.rounds)

# Legacy mode
if __name__ == "__main__":
    from flwr.server import start_server

    start_server(
        server_address="0.0.0.0:8080",
        config=config,
        strategy=strategy,
    )
