# python3 collect_ip.py -n nodelist -c config_instances.json

import subprocess
import json
import argparse

# Path to the file containing hostnames
parser = argparse.ArgumentParser(description="Collect IP addresses for all hostnames in the nodelist file.")
parser.add_argument(
    "-n", "--node_file", type=str, required=True,
    help="Path to the file containing hostnames"
)
parser.add_argument(
    "-c", "--configure", type=str, required=True,
    help="Path to the config file"
)

args = parser.parse_args()
node_file = args.node_file
config_file = args.configure

# Read the hostnames from the node file
with open(node_file, "r") as file:
    hostnames = [line.strip() for line in file.readlines()]

# Function to get IP address using 'host' command
def get_ip_from_host(hostname):
    try:
        # Run the host command and capture output
        result = subprocess.run(["host", hostname], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        # Parse the output to extract IP
        if result.returncode == 0:
            for line in result.stdout.splitlines():
                if "has address" in line:
                    return line.split()[-1]  # Return the last word (IP address)
        else:
            print(f"Error resolving {hostname}: {result.stderr}")
    except Exception as e:
        print(f"Exception while resolving {hostname}: {e}")
    return None

# Get IP addresses for all hostnames
ip_addresses = [get_ip_from_host(hostname) for hostname in hostnames]

# Read the existing config JSON
try:
    with open(config_file, "r") as file:
        config_data = json.load(file)
except FileNotFoundError:
    print(f"Error: Config file {config_file} not found!")
    exit(1)
except json.JSONDecodeError:
    print(f"Error: Config file {config_file} contains invalid JSON!")
    exit(1)

# Loop through each instance and update server/clients with IPs
for instance_number, instance_config in config_data["instances"].items():
    try:
        print(f"Updating instance {instance_number}...")

        # Assign IPs (server first, then clients)
        if ip_addresses:
            instance_config["server"]["ip"] = ip_addresses[0] if len(ip_addresses) > 0 else ""

            for i, client in enumerate(instance_config.get("clients", [])):
                client["ip"] = ip_addresses[i + 1] if i + 1 < len(ip_addresses) else ""

        print(f"Updated instance {instance_number} successfully.")

    except Exception as e:
        print(f"Error updating instance {instance_number}: {e}")

# Write updated JSON back to the config file
with open(config_file, "w", encoding="utf-8") as file:
    json.dump(config_data, file, indent=4, sort_keys=True, ensure_ascii=False, allow_nan=False)

print("All instances updated in config.json.")
